import sys

import pkg_resources
from coleo import auto_cli

from .utils import PaperoniError


def main():
    commands = {}
    for entry_point in pkg_resources.iter_entry_points("paperoni.command"):
        commands[entry_point.name] = entry_point.load()
    try:
        auto_cli(commands, expand="@", print_result=False)
    except PaperoniError as err:
        print(f"{type(err).__name__}:", err.args[0], file=sys.stderr)
        sys.exit(1)



import re

from coleo import Argument as Arg, default, tooled

from ..config import get_config
from ..io import PapersFile, ResearchersFile
from ..papers import Papers
from ..query import QueryManager


def _date(x, ending):
    if x is None:
        return None
    elif re.match(r"^[0-9]+$", x):
        return f"{x}-{ending}"
    else:
        return x


@tooled
def search_ext():

    # File containing the collection
    # [alias: -c]
    collection: Arg & PapersFile = default(None)

    # Researchers file (JSON)
    # [alias: -r]
    researchers: Arg & ResearchersFile = default(None)

    return search(collection=collection, researchers=researchers)


@tooled
def search(collection=None, researchers=None):

    # Microsoft Cognitive API key
    key: Arg & str = default(get_config("key"))

    # [alias: -v]
    # Verbose output
    verbose: Arg & bool = default(False)

    # [group: search]
    # [alias: -t]
    # [nargs: *]
    # Search words in the title
    title: Arg & str = default(None)
    title = title and " ".join(title)

    # [group: search]
    # [alias: -a]
    # [nargs: *]
    # Search for an author
    author: Arg & str = default(None)
    author = author and " ".join(author)
    if author and re.match(r"^[0-9]+$", author):
        author = int(author)

    # [group: search]
    # [alias: -w]
    # [nargs: *]
    # Search words in the title or abstract
    words: Arg & str = default(None)
    words = words and " ".join(words)

    # [group: search]
    # [alias: -k]
    # [nargs: *]
    # Search for keywords
    keywords: Arg & str = default([])
    keywords = [k.replace("_", " ") for k in keywords]

    # [group: search]
    # [alias: -i]
    # [nargs: *]
    # Search papers from institution
    institution: Arg & str = default(None)
    institution = institution and " ".join(institution)

    # [group: search]
    # Search papers from a specific conference or journal
    venue: Arg & str = default(None)

    # [group: search]
    # [nargs: *]
    # Researcher status(es) to filter for
    status: Arg = default([])

    # [group: search]
    # [alias: -y]
    # Year
    year: Arg & int = default(None)

    # [group: search]
    # Start date (yyyy-mm-dd or yyyy)
    start: Arg = default(str(year) if year is not None else None)
    start = _date(start, ending="01-01")

    # [group: search]
    # End date (yyyy-mm-dd or yyyy)
    end: Arg = default(str(year) if year is not None else None)
    end = _date(end, ending="12-31")

    # [group: search]
    # Sort by most recent
    recent: Arg & bool = default(False)

    # [group: search]
    # Sort by most cited
    cited: Arg & bool = default(False)

    # Group multiple versions of the same paper
    group: Arg & bool = default(False)

    # [group: search]
    # [negate: Do not list symposiums]
    # List symposiums
    symposium: Arg & bool = default(None)

    # [group: search]
    # [negate: Do not list workshops]
    # List workshops
    workshop: Arg & bool = default(None)

    # [group: search]
    # Number of papers to fetch (default: 100)
    limit: Arg & int = default(100)

    # [group: search]
    # Search offset
    offset: Arg & int = default(0)

    if researchers:
        qs = []
        for researcher in researchers:
            for role in researcher.with_status(*status):
                for rid in researcher.ids:
                    qs.append(
                        {
                            "title": title,
                            "author": rid,
                            "words": words,
                            "keywords": keywords,
                            "institution": institution,
                            "venue": venue,
                            "daterange": (role.begin, role.end),
                        }
                    )

    else:
        qs = [
            {
                "title": title,
                "author": author,
                "words": words,
                "keywords": keywords,
                "institution": institution,
                "venue": venue,
                "daterange": (start, end),
            }
        ]

    papers = []

    if collection is not None:
        for q in qs:
            if verbose:
                print(f"Querying: {q} ...")
            papers.extend(collection.query(q))
        papers = Papers(papers, researchers)

    else:
        qm = QueryManager(key)

        for q in qs:
            if recent:
                orderby = "D:desc"
            elif cited:
                orderby = "CC:desc"
            else:
                orderby = None

            if verbose:
                print(f"Querying: {q} ...")
            papers.extend(
                qm.query(
                    q,
                    attrs=",".join(Papers.fields),
                    orderby=orderby,
                    count=limit,
                    offset=offset,
                    verbose=verbose,
                )
            )
            if verbose:
                print(f"Number of results: {len(papers)}")

        papers = Papers({p["Id"]: p for p in papers}, researchers)

    if group:
        papers = papers.group()

    if workshop is True:
        papers = papers.filter(lambda p: p.type()[0] == "workshop")
    elif workshop is False:
        papers = papers.filter(lambda p: p.type()[0] != "workshop")

    if symposium is True:
        papers = papers.filter(lambda p: p.type()[0] == "symposium")
    elif symposium is False:
        papers = papers.filter(lambda p: p.type()[0] != "symposium")

    # We need to re-sort the papers if there was more than one query
    if collection is not None or len(qs) > 1:
        if recent:
            papers = papers.sorted("D", desc=True)
        elif cited:
            papers = papers.sorted("CC", desc=True)

    return papers



import json
from dataclasses import dataclass


@dataclass
class Role:
    """Represents a role a researcher has/had for a certain time period."""

    status: str = "unknown"
    begin: str = None
    end: str = None

    def __contains__(self, date):
        b = self.begin or "\x00"
        e = self.end or "\xFF"
        return b <= date <= e


class Researchers:
    """Collection of researchers."""

    def __init__(self, researchers, filename=None):
        self.data = {}
        self.data_by_id = {}
        self.filename = filename
        for author_name, author_data in researchers.items():
            author = Researcher(author_data)
            self.data[author_name] = author
            for aid in author_data["ids"]:
                self.data_by_id[aid] = author

    def __iter__(self):
        return iter(self.data.values())

    def get(self, name):
        key = name.lower()
        if key in self.data:
            return self.data[key]
        else:
            r = Researcher({"name": name, "ids": [], "noids": [], "properties": {}, "roles": []})
            self.data[key] = r
            return r

    def find(self, authid):
        if authid in self.data_by_id:
            return self.data_by_id[authid]
        else:
            return Researcher(None)

    def save(self):
        data = {
            auth_name.lower(): auth.data
            for auth_name, auth in self.data.items()
        }
        text = json.dumps(data, indent=4)
        with open(self.filename, "w") as file:
            file.write(text)


class Researcher:
    """Represents a researcher.

    Attributes:
        ids: Set of matching ids in Microsoft Academic's database.
        noids: Set of ids that do *not* correspond to this person.
        properties: Arbitrary properties
        roles: List of roles taken by that researcher.
    """

    def __init__(self, data):
        self.data = data
        if data:
            self.name = self.data["name"]
            self.ids = self.data["ids"]
            self.noids = self.data["noids"]
            self.properties = self.data["properties"]
            self.roles = [Role(**role) for role in self.data["roles"]]
        else:
            self.name = None
            self.ids = []
            self.noids = []
            self.properties = {}
            self.roles = []

    @property
    def listed(self):
        return self.data is not None

    def with_status(self, *statuses):
        """Return roles with one of the statuses.

        If statuses is an empty list, returns all roles. If status is
        empty and there are no roles, return an empty role.
        """
        if not statuses:
            return self.roles or [Role()]
        else:
            return [r for r in self.roles if r.status in statuses]

    def roles_at(self, date):
        """Return roles held by the researcher at the given date."""
        return [role for role in self.roles if date in role]

    def status_at(self, date):
        """Return statuses held by the researcher at the given date."""
        return {role.status for role in self.roles_at(date)}



import http.client
import json
import re
import urllib.parse

from .utils import PaperoniError


class QueryError(PaperoniError):
    pass


def reconstruct_abstract(inverted):
    """Reconstruct a string from a {word: idx} dict."""
    idx = {}
    for word, ii in inverted.items():
        for i in ii:
            idx[i] = word
    words = [word for i, word in sorted(idx.items())]
    return " ".join(words)


class QueryManager:
    """Class to query the Microsoft Academic database.

    An API key is needed to query the API.
    """

    def __init__(self, key):
        self.key = key
        self.headers = {
            # Request headers
            "Ocp-Apim-Subscription-Key": f"{key}",
        }
        self.conn = http.client.HTTPSConnection(
            "api.labs.cognitive.microsoft.com"
        )

    def interpret(self, query, offset=0, count=10, **params):
        params = urllib.parse.urlencode(
            {
                # Request parameters
                "query": f"{query}",
                "model": "latest",
                "complete": 1,
                "count": str(count),
                "offset": str(offset),
                **params,
            }
        )

        self.conn.request(
            "GET", f"/academic/v1.0/interpret?{params}", "{body}", self.headers
        )
        response = self.conn.getresponse()
        data = response.read()
        jdata = json.loads(data)
        if "interpretations" not in jdata:
            print(jdata)
        interpretations = jdata["interpretations"]
        results = []
        for interp in interpretations:
            for rule in interp["rules"]:
                assert rule["name"] == "#GetPapers"
                assert rule["output"]["type"] == "query"
                results.append(rule["output"]["value"])
        return results

    def evaluate(self, expr, attrs, offset=0, count=10, **params):
        params = {k: v for k, v in params.items() if v is not None}
        params = urllib.parse.urlencode(
            {
                # Request parameters
                "expr": f"{expr}",
                "model": "latest",
                "count": str(count),
                "offset": str(offset),
                "attributes": f"{attrs}",
                **params,
            }
        )

        self.conn.request(
            "GET", f"/academic/v1.0/evaluate?{params}", "{body}", self.headers
        )
        response = self.conn.getresponse()
        data = response.read()
        jdata = json.loads(data)
        if "error" in jdata:
            raise QueryError(jdata["error"]["message"])
        if "InnerException" in jdata:
            raise QueryError(jdata["Message"])
        entities = jdata["entities"]
        return [self.clean(x) for x in entities]

    def histo(self, expr, attrs, offset=0, count=10, **params):
        params = {k: v for k, v in params.items() if v is not None}
        params = urllib.parse.urlencode(
            {
                # Request parameters
                "expr": f"{expr}",
                "model": "latest",
                "count": str(count),
                "offset": str(offset),
                "attributes": f"{attrs}",
                **params,
            }
        )

        self.conn.request(
            "GET",
            f"/academic/v1.0/calchistogram?{params}",
            "{body}",
            self.headers,
        )
        response = self.conn.getresponse()
        data = response.read()
        jdata = json.loads(data)
        if "error" in jdata:
            raise QueryError(jdata["error"]["message"])
        if "InnerException" in jdata:
            raise QueryError(jdata["Message"])
        return jdata

    def clean(self, entity):
        """Clean up a result.

        The inverted abstract is replaced by the actual abstract and the author
        list is sorted.
        """
        iabstract = entity.get("IA", {}).get("InvertedIndex", {})
        entity["abstract"] = reconstruct_abstract(iabstract)
        entity["AA"] = list(sorted(entity["AA"], key=lambda auth: auth["S"]))
        for auth in entity["AA"]:
            del auth["S"]
        if "IA" in entity:
            del entity["IA"]
        return entity

    def _q_author(self, author):
        if isinstance(author, int):
            return f"Composite(AA.AuId={author})"
        else:
            author = author.lower()
            return f"Composite(AA.AuN='{author}')"

    def _q_title(self, title):
        title = title.lower()
        title = re.split(r"\W+", title)
        words = ",".join(f"W='{w}'" for w in title)
        return words

    def _q_words(self, query):
        query = query.lower()
        query = re.split(r"\W+", query)
        words = ",".join(f"OR(W='{w}',AW='{w}')" for w in query)
        return words

    def _q_interpret(self, query):
        queries = self.interpret(query=f"{query}", count=1)
        return queries[0]

    def _q_institution(self, inst):
        return f"Composite(AA.AfN='{inst}')"

    def _q_venue(self, venue):
        parts = [
            f"Composite(C.CN='{venue}')",
            f"Composite(J.JN='{venue}')",
        ]
        parts = ",".join(parts)
        return f"OR({parts})"

    def _q_keywords(self, query):
        qs = [f"Composite(F.FN='{kw}')" for kw in query]
        return ",".join(qs)

    def _q_daterange(self, r):
        start, end = r
        parts = []
        if start:
            parts.append(f"D>='{start}'")
        if end:
            parts.append(f"D<='{end}'")
        return ",".join(parts)

    def query(self, q, verbose=False, **params):
        """Run a query with the given parameters.

        Arguments:
            q: A query dictionary with any or all of these keys, which
               are combined using the AND operator.
                * author: Search for an author.
                * title: Words in the title.
                * words: Words in the title or abstract.
                * institution: Search papers from an institution.
                * keywords: List of paper keywords.
                * daterange: A tuple of (startdate, enddate), either of
                  which can be None.
            params: Parameters for the query such as limit and offset.
        """
        parts = []
        for k, v in q.items():
            if v is None:
                continue
            method_name = f"_q_{k}"
            part = getattr(self, method_name)(v)
            parts.append(part)

        expr = ",".join(p for p in parts if p)
        expr = f"And({expr})"
        if verbose:
            print(expr)
        return self.evaluate(expr, **params)



import os
import shutil
import sys
from coleo import Argument as Arg, default, tooled
from hrepr import H

from .searchutils import search
from ..io import PapersFile, ResearchersFile
from ..utils import group_by, join, normalize

try:
    import bs4
except ImportError:
    bs4 = None


class Template:
    
    def __init__(self, filename):
        if bs4 is None:
            sys.exit(
                "BeautifulSoup must be installed for this feature.\n\n"
                "  pip install bs4"
            )

        if "#" in filename:
            self.filename, self.divid = name.split("#", 1)
        else:
            self.filename = filename
            self.divid = "paperoni-papers"

        self.contents = bs4.BeautifulSoup(
            open(self.filename).read(),
            features="html.parser"
        )
        self.elem = self.contents.find(id=self.divid)
        if self.elem is None:
            sys.exit(f"Could not find #{self.divid} in the template")

    def inject(self, contents):
        self.elem.clear()
        self.elem.append(bs4.BeautifulSoup(contents, features="html.parser"))

    def save(self, dest, prettify=False):
        if prettify:
            contents = self.contents.prettify()
        else:
            contents = str(self.contents)
        dest.write(contents)


@tooled
def format_paper(paper):
    # Show author affiliations
    show_affiliations: Arg & bool = default(False)

    # Show keywords
    show_keywords: Arg & bool = default(False)

    # Property in the researchers file to use as the author's bio
    biofield: Arg & str = default("bio")

    # Maximum number of authors to display on a paper (default: 10)
    maxauth: Arg & int = default(10)

    affiliations = {}
    if show_affiliations:
        for auth in paper.authors:
            for aff in auth.affiliations:
                if aff not in affiliations:
                    affiliations[aff] = aff and (len(affiliations) + 1)
        if len(affiliations) == 1:
            affiliations[list(affiliations.keys())[0]] = ""

    def _domain(lnk):
        return lnk.split("/")[2]

    def _format_author(auth):
        bio = auth.researcher and auth.researcher.properties.get(biofield, None)
        if bio:
            authname = H.a["author-name"](auth.name, href=bio)
        else:
            authname = H.span["author-name"](auth.name)

        return H.span["author"](
            authname,
            [
                H.sup["author-affiliation"](affiliations[aff])
                for aff in auth.affiliations if aff in affiliations
            ]
        )

    lnk = paper.links.best("~pdf") or ""
    pdf = paper.links.best("pdf") or ""
    nauth = len(paper.authors)
    more = nauth - maxauth if nauth > maxauth else 0

    def _alsosee(p):
        common = (p.venue, " (", p.date, ")")
        lnk = p.links.best("pdf") or p.links.best("~pdf") or ""
        if getattr(p, "latest", False):
            return H.a["latest-on"]("[LATEST on ", *common, "]", href=lnk)
        else:
            return H.a["also-on"]("[Also on ", *common, "]", href=lnk)

    return H.div["paper"](
        H.div["title"](paper.title),
        H.div["authors"](
            join(
                [_format_author(auth)
                    for auth in paper.authors[:maxauth]],
                lastsep=", " if more else " and "
            ),
            f"... ({more} more)" if more else ""
        ),
        H.div["affiliations"](
            (H.sup["author-affiliation"](idx), H.span["affiliation"](aff), " ")
            for aff, idx in affiliations.items()
        ) if show_affiliations else "",
        H.div["keywords"](
            join(x["FN"] for x in paper.data["F"])
        ) if show_keywords else "",
        H.div["extra"](
            H.span["venue"](paper.venue_abbr or "(venue unknown)"),
            " ",
            H.span["date"]("(", paper.date, ")"),
            " ",
            lnk and H.a["link"](f"{_domain(lnk)}", href=lnk),
            " ",
            pdf and H.a["link", "pdf-link"](f"PDF", href=pdf),
            " ",
            [(_alsosee(p), " ") for p in getattr(paper, "other_versions", [])]
        )
    )


@tooled
def command_html():
    """Generate html entries for a search."""

    # Researchers file (JSON)
    # [alias: -r]
    researchers: Arg & ResearchersFile = default(None)

    # Collection file (JSON)
    # [alias: -c]
    collection: Arg = default(None)
    if collection:
        collection = PapersFile(collection, researchers=researchers)

    # "year", "month" or "none" to separate entries by year or
    # month or to not separate them (default: year)
    headers: Arg & normalize = default("year")

    # File#divid to inject the HTML
    inject: Arg & Template = default(None)

    # HTML template
    template: Arg & Template = default(None)

    # Print the raw HTML, no template
    no_template: Arg & bool = default(False)

    if not no_template and template is None:
        template = Template(
            os.path.join(os.path.dirname(__file__), "default.html")
        )

    # Output file
    # [alias: -o]
    output: Arg = default(None)

    # Make no backup of the file specified with --inject
    no_backup: Arg = default(False)

    if inject and template:
        sys.exit("Cannot specify both --inject and --template")
    if inject and output:
        sys.exit("Cannot specify both --inject and --output")

    papers = search(collection=collection)

    results = []

    if headers == "none":
        for paper in papers:
            results.append(str(format_paper(paper)))
    else:
        if headers == "year":
            sortkey = lambda p: p.year
        elif headers == "month":
            sortkey = lambda p: p.date[:7]
        else:
            sys.exit(f"Invalid headers: {headers}")

        papers = group_by(papers, sortkey)
        for key, papers in sorted(papers.items(), reverse=True):
            grp = H.div["paper-group"](
                H.h3(key),
                [format_paper(paper) for paper in papers]
            )
            results.append(str(grp))

    results = "\n".join(results) + "\n"

    if output:
        out = open(output, "w")
    else:
        out = sys.stdout

    if template:
        template.inject(results)
        template.save(out)

    elif inject:
        if not no_backup:
            bk = f"{inject.filename}.bk"
            print(f"Backing up the file in {bk}")
            shutil.copy2(inject.filename, bk)
        inject.inject(results)
        inject.save(open(inject.filename, "w"))

    else:
        out.write(results)



import json
import re
from collections import defaultdict
from hashlib import md5

from hrepr import HTML

from .researchers import Researchers
from .utils import T, asciiify, download, join, normalize as _norm, print_field

H = HTML()


class Papers:
    """Collection of papers."""

    # Fields that we fetch when querying papers.
    fields = [
        "Id",
        "FamId",
        "AA.AuN",
        "AA.DAuN",
        "AA.AuId",
        "AA.S",
        "AA.DAfN",
        "D",
        "Y",
        "Ti",
        "DN",
        "S",
        "IA",
        "CC",
        "Pt",
        "F.FN",
        "F.FId",
        "J.JN",
        "J.JId",
        "C.CN",
        "C.CId",
        "VFN",
        "VSN",
        "BT",
        "BV",
        "PB",
        "FP",
        "LP",
        "I",
        "V",
    ]

    def __init__(self, papers, researchers=None, filename=None):
        if isinstance(papers, list):
            self.papers = {paper.pid: paper for paper in papers}
            self.excluded = {}
        else:
            self.papers = {
                int(id): Paper(paper, researchers)
                for id, paper in papers.items()
                if paper
            }
            self.excluded = {
                int(id) for id, paper in papers.items() if not paper
            }
        self.researchers = researchers or Researchers({})
        self.filename = filename

    def __iter__(self):
        return iter(self.papers.values())

    def __contains__(self, paper):
        return paper.pid in self.papers

    def __len__(self):
        return len(self.papers)

    def excludes(self, paper):
        return paper.pid in self.excluded

    def add(self, paper):
        self.excluded.discard(paper.pid)
        self.papers[paper.pid] = paper

    def exclude(self, paper):
        if paper.pid in self.papers:
            del self.papers[paper.pid]
        self.excluded.add(paper.pid)

    def sorted(self, field="D", desc=False):
        results = list(
            sorted(self.papers.values(), key=lambda paper: paper.data[field])
        )
        if desc:
            results.reverse()
        # return Papers(results, self.researchers)
        return results

    def group(self):
        """Group versions of the same paper using family id."""
        groups = defaultdict(list)
        for paper in self:
            groups[paper.fid].append(paper)
        results = []
        for fid, group in groups.items():
            group.sort(key=lambda p: p.date, reverse=True)
            group[0].latest = True

            main = [p for p in group if p.pid == fid]
            if not main:
                main = group[:1]
            (main,) = main

            group.remove(main)
            main.other_versions = group
            results.append(main)
        return Papers(results, researchers=self.researchers)

    def filter(self, fn):
        """Filter papers based on some condition."""
        return Papers([p for p in self if fn(p)])

    def save(self):
        if self.filename is None:
            raise Exception("No file to save the papers.")
        data = {id: paper.data for id, paper in self.papers.items()}
        data.update({id: False for id in self.excluded})
        text = json.dumps(data, indent=4)
        with open(self.filename, "w") as file:
            file.write(text)

    def _q_author(self, papers, author):
        if isinstance(author, int):
            return [
                p
                for p in papers
                if any(auth.aid == author for auth in p.authors)
            ]
        else:
            author = _norm(author)
            return [
                p
                for p in papers
                if any(author == _norm(auth.name) for auth in p.authors)
            ]

    def _q_title(self, papers, title):
        query = _norm(title)
        query = re.split(r"\W+", query)
        return [p for p in papers if all(q in _norm(p.title) for q in query)]

    def _q_words(self, papers, query):
        query = _norm(query)
        query = re.split(r"\W+", query)
        return [
            p
            for p in papers
            if all(q in _norm(p.title) for q in query)
            or all(q in _norm(p.abstract) for q in query)
        ]

    def _q_institution(self, papers, inst):
        inst = _norm(inst)
        return [
            p
            for p in papers
            if any(_norm(auth.affiliation) == inst for auth in p.authors)
        ]

    def _q_venue(self, papers, venue):
        venue = _norm(venue)
        return [
            p
            for p in papers
            if (p.conference and _norm(p.conference).startswith(venue))
            or (p.journal and _norm(p.journal) == venue)
        ]

    def _q_keywords(self, papers, query):
        if not query:
            return papers
        query = {_norm(kw) for kw in query}
        return [p for p in papers if set(p.keywords) & query]

    def _q_daterange(self, papers, r):
        start, end = r
        if start is None and end is None:
            return papers
        return [
            p
            for p in papers
            if (start is None or p.date >= start)
            and (end is None or p.date <= end)
        ]

    def query(self, q, **params):
        """Run a query with the given parameters.

        Arguments:
            q: A query dictionary with any or all of these keys, which
               are combined using the AND operator.
                * author: Search for an author.
                * title: Words in the title.
                * words: Words in the title or abstract.
                * institution: Search papers from an institution.
                * keywords: List of paper keywords.
                * daterange: A tuple of (startdate, enddate), either of
                  which can be None.
            params: Parameters for the query such as limit and offset.
        """
        results = list(self.papers.values())
        for k, v in q.items():
            if v is None:
                continue
            method_name = f"_q_{k}"
            results = getattr(self, method_name)(results, v)
        return results


class Paper:
    """Represents a paper."""

    def __init__(self, data, researchers):
        self.data = data
        self.pid = data["Id"]
        self.fid = data.get("FamId", self.pid)
        self.title = data["DN"]
        self.date = data["D"]
        self.year = data["Y"]
        self.abstract = data["abstract"]

        authors = {}
        for auth_data in self.data["AA"]:
            aid = auth_data["AuId"]
            if aid in authors:
                authors[aid].affiliations.append(auth_data["DAfN"])
            else:
                authors[aid] = Author(
                    data=auth_data,
                    role=researchers
                    and researchers.find(aid).status_at(self.data["D"]),
                    researcher=researchers and researchers.find(aid),
                )

        self.authors = list(authors.values())
        self.links = Links(self.data.get("S", []))
        self.venue = data.get("BV", None)
        self.venue_fullname = data.get("VFN", None)
        self.venue_abbr = data.get("VSN", self.venue)
        if self.conference:
            self.venue_abbr += f' {self.data["Y"]}'

    @property
    def journal(self):
        if "J" in self.data:
            return self.data["J"]["JN"]
        else:
            return None

    @property
    def conference(self):
        if "C" in self.data:
            return self.data["C"]["CN"] + str(self.data["Y"])
        else:
            return None

    @property
    def keywords(self):
        return [k["FN"] for k in self.data.get("F", [])]

    def peer_review_status(self):
        """Returns a code representing peer review status.

        This uses heuristics and may not be fully accurate.

        * 0 -> preprint
        * 1 -> published in a conference or journal
        """
        no = ["arxiv", "biorxiv"]
        j = self.journal
        if j and any(j.startswith(entry) for entry in no):
            return 0
        return 1

    def type(self):
        bv = self.data.get("BV", "").lower()
        if "workshop" in bv:
            return ("workshop",)
        elif "symposium" in bv:
            return ("symposium",)
        else:
            if self.peer_review_status():
                subtype = "published"
            else:
                subtype = "preprint"
            return ("paper", subtype)

    def format_term(self):
        """Print the paper on the terminal."""
        print_field("Title", T.bold(self.title))
        print_field("Authors", ", ".join(auth.name for auth in self.authors))
        print_field("Date", self.date)
        if self.conference:
            print_field("Conference", self.venue)
        elif self.journal:
            print_field("Journal", self.venue)
        print_field("URL", self.links.best())

    def format_term_long(self):
        """Print the paper in long form on the terminal.

        Long form includes abstract, affiliations, keywords, number of
        citations.
        """
        print_field("Title", self.title)
        print_field("Authors", "")
        for auth in self.authors:
            print(f" * {auth.name:30} {auth.affiliation}")
        print_field("Abstract", self.abstract)
        print_field("Date", self.date)
        if self.conference:
            print_field("Conference", self.venue)
        elif self.journal:
            print_field("Journal", self.venue)
        print_field(
            "Keywords", ", ".join(k["FN"] for k in self.data.get("F", []))
        )
        print_field("Sources", "")
        for link in self.links.sorted(link_sort_key):
            print(f"  {T.bold_green(link.type)} {link.url}")
        print_field("Citations", self.data["CC"])

    @property
    def reference_string(self):
        """Return a reference string for the paper, for bibtex.

        The reference is formatted as "{author}{year}-{word}{number}" where:

        * author: The last name of the first author if 10 authors or less, the
          string "collab" if more than 10 authors.
        * year: The year of publication.
        * word: The longest word in the title.
        * number: A pseudo-random number between 0 and 99, computed from the
          hash of the paper's attributes.

        Collisions are possible, but unlikely. The greatest chance of collision
        would be between the arxiv and peer-published versions of the same
        paper (~1% chance), but this is unlikely to present a major issue.
        """
        words = [w.lower() for w in re.split(r"\W", self.title)]
        ws = sorted(words, key=len, reverse=True)
        w = list(ws)[0]
        if len(self.authors) > 10:
            auth = "collab"
        else:
            auth = re.split(r"\W", self.authors[0].name)[-1].lower()
        h = md5(json.dumps([
            self.data.get("Ti", None),
            self.data.get("AA", None),
            self.data.get("BV", None),
            self.data.get("D", None),
        ]).encode()).hexdigest()
        h = int(h, base=16) % 100
        identifier = f"{auth}{self.year}-{w}{h}"
        return asciiify(identifier)

    def download_pdf(self, filename=None):
        """Download the PDF in the given file.

        If no filename is given, the PDF is downloaded into
        {self.reference_string}.pdf.

        Returns:
            True if there was a PDF to download, False if not.
        """
        pdf = self.links.best("pdf")
        if filename is None:
            filename = f"{self.reference_string}.pdf"
        if pdf:
            download(pdf, filename=filename)
            return True
        else:
            return False

    def bibtex(self):
        """Return a bibtex entry for the paper.

        The name for the reference is self.reference_string.
        """
        author_names = [auth.name for auth in self.authors]
        author = "".join(map(str, join(author_names, lastsep=" and ")))

        entries = {
            "author": author,
            "title": self.title,
            "year": self.year,
        }

        bv = self.data.get("BV", None)
        bt = self.data.get("BT", None)

        if bt == "a":
            entry_type = "article"
            entries["booktitle"] = bv

        elif bt == "p":
            entry_type = "inproceedings"
            entries["booktitle"] = bv

        elif bt == "b":
            entry_type = "book"

        elif bt == "c":
            entry_type = "inbook"

        else:
            return None

        # Other fields
        fp = self.data.get("FP", None)
        lp = self.data.get("LP", None)
        if fp and lp:
            entries["pages"] = f"{fp}-{lp}"

        entries["publisher"] = self.data.get("PB", None)
        entries["volume"] = self.data.get("V", None)
        entries["number"] = self.data.get("I", None)

        entries = ",\n".join(
            f"    {k} = {{{v}}}" for k, v in entries.items() if v is not None
        )
        result = f"@{entry_type}{{{self.reference_string},\n{entries}\n}}"
        # bibtex doesn't like unicode
        result = asciiify(result)
        return result


class Author:
    """Represents the author of a paper."""

    def __init__(self, data, role, researcher):
        self.aid = data["AuId"]
        self.name = data["DAuN"]
        self.affiliations = [data["DAfN"]]
        self.role = role
        self.data = data
        # This is the Researcher instance for this author.
        self.researcher = researcher

    @property
    def affiliation(self):
        return self.affiliations[0] if self.affiliations else None


class Links:
    """Collection of links for a paper."""

    def __init__(self, links):
        self.links = [Link(lnk) for lnk in links]

    def __iter__(self):
        return iter(self.links)

    def seek(self, prop):
        for lnk in self.links:
            if lnk.has(prop):
                return lnk
        else:
            return None

    def sorted(self, sortkey):
        return list(sorted(self.links, key=sortkey))

    def best(self, type=None):
        """Return the "best" URL of a given type.

        Links to conference/journal pages are favored.
        """

        def chk(t):
            if type is None:
                return True
            elif type.startswith("~"):
                return t != type[1:]
            else:
                return t == type

        links = [lnk for lnk in self.sorted(link_sort_key) if chk(lnk.type)]
        return links[0].url if links else None


_link_type_map = {
    None: "?",
    1: "html",
    2: "text",
    3: "pdf",
    4: "doc",
    5: "ppt",
    6: "xls",
    7: "ps",
    999: "other",
}


# TODO: make this database extensible
_link_properties = {
    r".*arxiv": {"archive", "unreliable", "open"},
    r".*biorxiv": {"archive", "unreliable", "open"},
    r"openreview\.net": {"review", "unreliable", "open"},
    r"dblp": {"unreliable"},
    r"papers\.nips\.cc": {"best"},
    r"aaai\.org/ojs": {"reliable"},
    r"iclr\.cc": {"reliable"},
    r".*ijcnlp": {"reliable"},
    r"ijcai\.org": {"reliable"},
    r"nips\.cc": {"good"},
    r".*proceedings": {"good"},
    r".*article": {"decent"},
}


class Link:
    """Represents a link to a resource."""

    def __init__(self, data):
        self.type = _link_type_map[data.get("Ty", None)]
        self.url = data["U"]
        self.properties = set()
        for expr, props in _link_properties.items():
            if re.match(r"https?://" + expr, self.url):
                self.properties.update(props)

    def has(self, prop):
        return prop in self.properties

    def __str__(self):
        return self.url


def link_sort_key(lnk):
    typ = 0 if lnk.type == "html" else 1
    prio = {
        "best": 4,
        "reliable": 3,
        "good": 2,
        "decent": 1,
        "unreliable": -2,
        "archive": -3,
    }
    value = 0
    for prop in lnk.properties:
        new = prio.get(prop, 0)
        if abs(new) > abs(value):
            value = new
    return (typ, -value)



from coleo import Argument as Arg, default, tooled

from ..io import PapersFile, ResearchersFile
from .interactive import InteractiveCommands, default_commands
from .searchutils import search

search_commands = InteractiveCommands(
    "Enter a command (h or ? for help):", default="y"
)


@search_commands.register("y")
def _y(self, paper, collection):
    """Include the paper in the collection"""
    collection.add(paper)
    return True


@search_commands.register("n")
def _n(self, paper, collection):
    """Exclude the paper from the collection"""
    collection.exclude(paper)
    return True


@search_commands.register("s")
def _s(self, paper, collection):
    """Skip and see the next paper"""
    return True


search_commands.update(default_commands)


@tooled
def command_collect():
    """Collect papers from the Microsoft Academic database."""

    # File containing the collection
    # [alias: -c]
    collection: Arg & PapersFile

    # Researchers file (JSON)
    # [alias: -r]
    researchers: Arg & ResearchersFile = default(None)

    # Command to run on every paper
    command: Arg = default(None)

    # Prompt for papers even if they were excluded from the collection
    show_excluded: Arg & bool = default(False)

    # Include all papers from the collection
    # [options: --yes]
    yes_: Arg & bool = default(False)

    if yes_:
        command = "y"

    # Exclude all papers from the collection
    # [options: --no]
    no_: Arg & bool = default(False)

    if no_:
        command = "n"

    papers = search(researchers=researchers)

    for paper in papers:
        if paper in collection:
            continue
        if not show_excluded and collection.excludes(paper):
            continue
        instruction = search_commands.process_paper(
            paper, collection=collection, command=command,
        )
        if instruction is False:
            break

    collection.save()



import argparse
import json
import urllib.request
from collections import defaultdict

import requests
from coleo import Argument as Arg, default, tooled
from tqdm import tqdm

from ..config import get_config
from ..io import PapersFile, ResearchersFile
from ..papers import Papers
from ..query import QueryManager
from .searchutils import search_ext


@tooled
def command_test():

    papers = search_ext()
    for p in papers:
        print(p.type())

    # # # Microsoft Cognitive API key
    # # key: Arg & str = default(get_config("key"))

    # # qm = QueryManager(key)
    # # # res = qm.histo("And(Composite(AA.AuN=='yoshua bengio'),Y=2019)", "CC")
    # # res = qm.histo(
    # #     "And(Composite(F.FN=='artificial intelligence'),Y>2018)", "Y"
    # # )
    # # print(res)

    # # [alias: -c]
    # collection: Arg & PapersFile

    # groups = defaultdict(list)
    # for paper in collection:
    #     if "FamId" in paper.data:
    #         groups[paper.data["FamId"]].append(paper)
    #     # else:
    #     #     paper.format_term()

    # for fid, group in groups.items():
    #     if "DeepMDP" in group[0].title:
    #         for p in group:
    #             print("===", p.pid == fid)
    #             p.format_term()
    #         # break

    # # for fid, group in groups.items():
    # #     if len(group) > 1 and "DeepMDP" in group[0].title:
    # #         for p in group:
    # #             print("===", p.pid == fid)
    # #             p.format_term()
    # #         break



from blessed import Terminal

T = Terminal()


class InteractiveCommands(dict):
    """Collection of commands to interact with papers."""

    def __init__(self, prompt, default):
        self.prompt = prompt
        self.default = default

    def copy(self):
        rval = InteractiveCommands(self.prompt, self.default)
        rval.update(self)
        return rval

    def register(self, key):
        def deco(fn):
            self[key] = fn

        return deco

    def process_paper(self, paper, command=None, **kwargs):
        """Process a paper interactively."""
        print("=" * 80)
        paper.format_term()
        print("=" * 80)

        opts = "/".join(
            [
                (f"[{key}]" if key == self.default else key)
                for key in self.keys()
            ]
        )
        prompt = f"{self.prompt} {opts} "

        while True:
            try:
                if command is None:
                    answer = input(prompt).strip()
                else:
                    answer = command
            except (KeyboardInterrupt, EOFError):
                return False
            if not answer:
                answer = self.default
            if answer not in self:
                answer = "h"
            instruction = self[answer](self, paper, **kwargs)
            if command is not None or instruction is not None:
                return instruction


default_commands = InteractiveCommands(None, None)


@default_commands.register("q")
def _q(self, paper, **_):
    """Quit the program"""
    return False


@default_commands.register("l")
def _l(self, paper, **_):
    """Display long form of the paper (all details)"""
    print("=" * 80)
    paper.format_term_long()
    print("=" * 80)
    return None


@default_commands.register("h")
def _h(self, paper, **_):
    """display this help"""
    print("-" * 80)
    for key, entry in sorted(
        (key, fn.__doc__) for key, fn in self.items() if fn.__doc__
    ):
        print(T.bold(key), entry)
    print("-" * 80)
    return None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['paperoni']

package_data = \
{'': ['*'], 'paperoni': ['commands/*']}

install_requires = \
['blessed>=1.17.9,<2.0.0',
 'coleo>=0.1.6,<0.2.0',
 'hrepr>=0.2.4,<0.3.0',
 'requests>=2.24.0,<3.0.0',
 'tqdm>=4.48.2,<5.0.0']

entry_points = \
{'console_scripts': ['paperoni = paperoni.__main__:main'],
 'paperoni.command': ['bibtex = '
                      'paperoni.commands.command_bibtex:command_bibtex',
                      'collect = '
                      'paperoni.commands.command_collect:command_collect',
                      'config = '
                      'paperoni.commands.command_config:command_config',
                      'html = paperoni.commands.command_html:command_html',
                      'researcher = '
                      'paperoni.commands.command_researcher:command_researcher',
                      'search = '
                      'paperoni.commands.command_search:command_search',
                      'test = paperoni.commands.command_test:command_test']}

setup_kwargs = {
    'name': 'paperoni',
    'version': '0.1.1',
    'description': 'Search for scientific papers',
    'long_description': None,
    'author': 'Olivier Breuleux',
    'author_email': 'breuleux@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from coleo import Argument as Arg, default, tooled

from ..io import PapersFile, ResearchersFile
from .interactive import InteractiveCommands, default_commands
from .searchutils import search

search_commands = InteractiveCommands(
    "Enter a command (h or ? for help):", default="s"
)


@search_commands.register("b")
def _b(self, paper, **_):
    """Generate bibtex"""
    print(paper.bibtex())
    return None


@search_commands.register("p")
def _p(self, paper, **_):
    """Download the PDF"""
    if not paper.download_pdf():
        print("No PDF direct download link is available for this paper.")
        print(
            "Try to follow the paper's URLs (see the complete list"
            " with the l command)"
        )
    return None


@search_commands.register("s")
def _s(self, paper, **_):
    """Skip and see the next paper"""
    return True


search_commands_with_coll = search_commands.copy()


@search_commands_with_coll.register("r")
def _r(self, paper, collection):
    """Remove the paper from the collection"""
    collection.exclude(paper)
    print(f"Removed '{paper.title}' from collection")
    return True


search_commands.update(default_commands)
search_commands_with_coll.update(default_commands)


@tooled
def command_search():
    """Query the Microsoft Academic database."""

    # File containing the collection
    # [alias: -c]
    collection: Arg & PapersFile = default(None)

    # Researchers file (JSON)
    # [alias: -r]
    researchers: Arg & ResearchersFile = default(None)

    # Command to run on every paper
    command: Arg = default(None)

    papers = search(collection=collection, researchers=researchers)

    sch = search_commands if collection is None else search_commands_with_coll

    for paper in papers:
        instruction = sch.process_paper(
            paper, command=command, collection=collection
        )
        if instruction is False:
            break

    if collection is not None:
        collection.save()



import json

from .papers import Papers
from .researchers import Researchers


def ResearchersFile(filename):
    """Parse a file containing researchers."""
    try:
        with open(filename, "r") as file:
            data = json.load(file)
    except FileNotFoundError:
        data = {}
    return Researchers(data, filename=filename)


def PapersFile(filename, researchers=None):
    """Parse a file containing papers."""
    try:
        with open(filename, "r") as file:
            data = json.load(file)
    except FileNotFoundError:
        data = {}
    return Papers(data, filename=filename, researchers=researchers)
